require 'metasm'
include Metasm
require 'optparse'
require 'deobfuscate'
require 'analyse_handlers'
require 'macro_cpu_vm'
require 'decompile'

# parse arguments
opts = {}
OptionParser.new { |opt|
	opt.banner = './go.rb [options] [<1stvmhandler> [<1stvminstr>]]'
	opt.on('--no-data', 'do not display data bytes') { opts[:nodata] = true }
	opt.on('--no-data-trace', 'do not backtrace memory read/write accesses') { opts[:nodatatrace] = true }
	opt.on('-v', '--verbose') { $VERBOSE = true }
	opt.on('-d', '--debug') { $DEBUG = true }
	opt.on('--nodeobf') { opts[:nodeobf] = true }
	opt.on('--not2macro') { opts[:not2macro] = true }
	opt.on('--dump-driver') { opts[:dumpdriver] = true }
	opt.on('--dump-data') { opts[:dumpdata] = true }
	opt.on('--nosave') { File.unlink('saved-handlers') if File.exist?('saved-handlers') }
	opt.on('--showrdtsc') { opts[:showrdtsc] = true }
}.parse!(ARGV)

exe = AutoExe.decode_file 'driver.sys'
exe.cpu = (opts[:nodeobf] ? Ia32 : Ia32_junkfilter).new
t2exe = exe.dup
t2exe.cpu = (opts[:not2macro] ? T2CPU : T2MacroCPU).new(exe)
t2exe.cpu.ia32dasm.cpu.showrdtsc = true if opts[:showrdtsc]
t2dasm = T2Disassembler.new(t2exe)

if not ARGV.empty?
	t2entry = T2Addr.new(ARGV.shift.to_i(16), (ARGV.shift || '0').to_i(16))
	t2dasm.auto_label_at t2entry, 'vmentry'
else
	t2entry = T2Addr.new(0x13000+0xe9fe, 0x13000+0xeea6)
end
#t2dasm.decoded[T2Addr.new(0x144cc, 0x3807d)] = true
h = t2dasm.address_binding[t2entry] = {}
h[:init_esp] = h[:esp] = Expression['ctx', :+, 66000]
h[:host_data] = Expression[0x13000]
h[:ctx_256] = Expression['ctx', :+, 256]
h[:host_esp] = Expression['host_esp']
h[:zero_20] = Expression[0]
h[:base_64_2c] = Expression[0x13000+0x40]


puts "disassembling t2vm from #{t2entry}"
time_start = Time.now
begin
	t2dasm.disassemble t2entry
	if opts[:dumpdriver]
		t2dasm.cpu.ia32dasm.parse_c_file 'driver.h'
		exe.disassemble(0x11a20, 0x11910, 0x11940, 0x11960, 0x119b0, 0x119d0, 0x3a520, 'entrypoint')
	end
rescue Exception
	puts $!, $!.message, $!.backtrace
end
time_diff = Time.now - time_start
puts "dasm took #{'%.2f' % time_diff}s"

puts 'handlers.asm'
File.open('handlers.asm', 'w') { |fd|
	t2dasm.cpu.ia32dasm.dump(opts[:dumpdata]) { |l| fd.puts l }
}

begin
	puts 'vm.c'
	File.open('vm.c', 'w') { |fd|
		fd.puts t2dasm.decompile(t2entry)
	}
rescue Exception
	puts $!, $!.message, $!.backtrace
end

puts 'vm.asm'
File.open('vm.asm', 'w') { |fd|
	na = t2dasm.normalize t2entry
	todo = [na]
	done = []
	while a = todo.pop
		next if done.include? a
		next if not b = t2dasm.decoded[a] or not b.kind_of? DecodedInstruction
		b = b.block
		done << a
		
		fd.puts "//stop#{" - back to #{na}" if na}" if na != b.address
		
		t2dasm.dump_block(b) { |l| fd.puts l }
		
		na = b.list.last.next_addr
		has_na = false
		b.each_to { |t, type|
			t = t2dasm.normalize t
			na = t if type == :subfuncret and b.list.last.instruction.opname == 'call'
			if t2dasm.function[t]: todo.unshift t
			elsif na == t: has_na = true
			else todo << t
			end
		}
		todo << na if has_na    # ensure this addr is next in the dump
	end
	fd.puts "//stop#{" - back to #{na}" if na}"
}
